@@ -1,4 +1,3 @@
-
 b_file_keyrings <- new.env(parent = emptyenv())
 
 #' Encrypted file keyring backend
@@ -12,7 +11,6 @@ b_file_keyrings <- new.env(parent = emptyenv())
 #'
 #' @family keyring backends
 #' @export
-#' @include backend-class.R
 #' @examples
 #' \dontrun{
 #' kb <- backend_file$new()
@@ -23,19 +21,28 @@ backend_file <- R6Class(
   inherit = backend_keyrings,
   public = list(
     name = "file",
-    initialize = function(keyring = NULL)
-      b_file_init(self, private, keyring),
+    initialize = function(keyring = NULL) b_file_init(self, private, keyring),
 
     get = function(service, username = NULL, keyring = NULL)
       b_file_get(self, private, service, username, keyring),
     get_raw = function(service, username = NULL, keyring = NULL)
       b_file_get_raw(self, private, service, username, keyring),
-    set = function(service, username = NULL, keyring = NULL)
-      b_file_set(self, private, service, username, keyring),
-    set_with_value = function(service, username = NULL, password = NULL,
-      keyring = NULL)
-      b_file_set_with_value(self, private, service, username, password,
-                            keyring),
+    set = function(service, username = NULL, keyring = NULL, prompt = NULL)
+      b_file_set(self, private, service, username, keyring, prompt),
+    set_with_value = function(
+      service,
+      username = NULL,
+      password = NULL,
+      keyring = NULL
+    )
+      b_file_set_with_value(
+        self,
+        private,
+        service,
+        username,
+        password,
+        keyring
+      ),
 
     delete = function(service, username = NULL, keyring = NULL)
       b_file_delete(self, private, service, username, keyring),
@@ -53,35 +60,45 @@ backend_file <- R6Class(
       b_file_keyring_unlock(self, private, keyring, password),
     keyring_is_locked = function(keyring = NULL)
       b_file_keyring_is_locked(self, private, keyring),
-    keyring_list = function()
-      b_file_keyring_list(self, private),
+    keyring_list = function() b_file_keyring_list(self, private),
 
-    keyring_default = function()
-      b_file_keyring_default(self, private),
+    keyring_default = function() b_file_keyring_default(self, private),
     keyring_set_default = function(keyring)
       b_file_keyring_set_default(self, private, keyring),
 
     docs = function() {
-      modifyList(super$docs(), list(. = paste0(
-        "Store secrets in encrypted files.\n",
-        private$keyring)))
+      modifyList(
+        super$docs(),
+        list(
+          . = paste0(
+            "Store secrets in encrypted files.\n",
+            private$keyring
+          )
+        )
+      )
     }
   ),
 
   private = list(
     keyring = NULL,
 
-    keyring_create_direct = function(keyring = NULL, password = NULL, prompt = NULL)
-      b__file_keyring_create_direct(self, private, keyring, password, prompt),
+    keyring_create_direct = function(
+      keyring = NULL,
+      password = NULL,
+      prompt = NULL
+    ) b__file_keyring_create_direct(self, private, keyring, password, prompt),
     keyring_autocreate = function(keyring = NULL)
       b__file_keyring_autocreate(self, private, keyring),
     keyring_file = function(keyring = NULL)
       b__file_keyring_file(self, private, keyring),
     keyring_read_file = function(keyring = NULL)
       b__file_keyring_read_file(self, private, keyring),
-    keyring_write_file = function(keyring = NULL, nonce = NULL, items = NULL,
-      key = NULL)
-      b__file_keyring_write_file(self, private, keyring, nonce, items, key),
+    keyring_write_file = function(
+      keyring = NULL,
+      nonce = NULL,
+      items = NULL,
+      key = NULL
+    ) b__file_keyring_write_file(self, private, keyring, nonce, items, key),
 
     get_keyring_pass = function(keyring = NULL)
       b__file_get_keyring_pass(self, private, keyring),
@@ -92,9 +109,12 @@ backend_file <- R6Class(
     is_set_keyring_pass = function(keyring = NULL)
       b__file_is_set_keyring_pass(self, private, keyring),
 
-    update_cache = function(keyring = NULL, nonce = NULL, check = NULL,
-      items = NULL)
-      b__file_update_cache(self, private, keyring, nonce, check, items),
+    update_cache = function(
+      keyring = NULL,
+      nonce = NULL,
+      check = NULL,
+      items = NULL
+    ) b__file_update_cache(self, private, keyring, nonce, check, items),
     get_cache = function(keyring = NULL)
       b__file_get_cache(self, private, keyring)
   )
@@ -106,7 +126,6 @@ b_file_init <- function(self, private, keyring) {
 }
 
 b_file_get <- function(self, private, service, username, keyring) {
-
   private$keyring_autocreate(keyring)
 
   username <- username %||% getOption("keyring_username")
@@ -118,13 +137,19 @@ b_file_get <- function(self, private, service, username, keyring) {
   item_matches <- all_services %in% service
 
   if (!is.null(username)) {
-    all_users <- vapply(all_items, function(x) x$user_name %||% NA_character_, character(1L))
+    all_users <- vapply(
+      all_items,
+      function(x) x$user_name %||% NA_character_,
+      character(1L)
+    )
     item_matches <- item_matches & all_users %in% username
   }
 
   if (sum(item_matches) < 1L) {
-    b_file_error("cannot get secret",
-                 "The specified item could not be found in the keychain.")
+    b_file_error(
+      "cannot get secret",
+      "The specified item could not be found in the keychain."
+    )
   }
 
   vapply(
@@ -136,23 +161,42 @@ b_file_get <- function(self, private, service, username, keyring) {
   )
 }
 
-b_file_set <- function(self, private, service, username, keyring) {
-
-  private$keyring_autocreate(keyring)
-
+b_file_set <- function(self, private, service, username, keyring, prompt) {
   username <- username %||% getOption("keyring_username")
-  if (self$keyring_is_locked(keyring)) self$keyring_unlock(keyring)
 
-  password <- get_pass()
+  keyring <- keyring %||% private$keyring
+  file <- private$keyring_file(keyring)
+  ex <- file.exists(file)
+
+  # We use a different prompt in this case, to give a heads up
+  prompt <- prompt %||%
+    if (!ex && interactive()) {
+      paste0(
+        "Note: the specified keyring does not exist, you'll have to ",
+        "create it in the next step. Key password: "
+      )
+    } else {
+      "Password: "
+    }
+
+  password <- get_pass(prompt)
+  if (is.null(password)) stop("Aborted setting keyring key")
+
+  private$keyring_autocreate()
 
   self$set_with_value(service, username, password, keyring)
 
   invisible(self)
 }
 
-b_file_set_with_value <- function(self, private, service, username,
-                                  password, keyring) {
-
+b_file_set_with_value <- function(
+  self,
+  private,
+  service,
+  username,
+  password,
+  keyring
+) {
   private$keyring_autocreate(keyring)
 
   username <- username %||% getOption("keyring_username")
@@ -166,7 +210,11 @@ b_file_set_with_value <- function(self, private, service, username,
     all_items <- cached$items
 
     services <- vapply(all_items, `[[`, character(1L), "service_name")
-    users <- vapply(all_items, function(x) x$user_name %||% NA_character_, character(1))
+    users <- vapply(
+      all_items,
+      function(x) x$user_name %||% NA_character_,
+      character(1)
+    )
     existing <- if (!is.null(username)) {
       services %in% service & users %in% username
     } else {
@@ -196,7 +244,6 @@ b_file_set_with_value <- function(self, private, service, username,
 }
 
 b_file_delete <- function(self, private, service, username, keyring) {
-
   username <- username %||% getOption("keyring_username")
   if (self$keyring_is_locked(keyring)) self$keyring_unlock(keyring)
 
@@ -208,7 +255,11 @@ b_file_delete <- function(self, private, service, username, keyring) {
     all_items <- cached$items
 
     services <- vapply(all_items, `[[`, character(1L), "service_name")
-    users <- vapply(all_items, function(x) x$user_name %||% NA_character_, character(1))
+    users <- vapply(
+      all_items,
+      function(x) x$user_name %||% NA_character_,
+      character(1)
+    )
     existing <- if (!is.null(username)) {
       services %in% service & users %in% username
     } else {
@@ -229,7 +280,6 @@ b_file_delete <- function(self, private, service, username, keyring) {
 }
 
 b_file_list <- function(self, private, service, keyring) {
-
   private$keyring_autocreate(keyring)
 
   cached <- private$get_cache(keyring)
@@ -237,7 +287,11 @@ b_file_list <- function(self, private, service, keyring) {
 
   res <- data.frame(
     service = vapply(all_items, `[[`, character(1L), "service_name"),
-    username = vapply(all_items, function(x) x$user_name %||% NA_character_, character(1)),
+    username = vapply(
+      all_items,
+      function(x) x$user_name %||% NA_character_,
+      character(1)
+    ),
     stringsAsFactors = FALSE
   )
 
@@ -253,7 +307,6 @@ b_file_keyring_create <- function(self, private, keyring, password) {
 }
 
 b_file_keyring_delete <- function(self, private, keyring) {
-
   self$keyring_lock(keyring)
 
   kr_file <- private$keyring_file(keyring)
@@ -264,13 +317,16 @@ b_file_keyring_delete <- function(self, private, keyring) {
 }
 
 b_file_keyring_lock <- function(self, private, keyring) {
-  private$keyring_autocreate(keyring)
+  keyring <- keyring %||% private$keyring
+  file <- private$keyring_file(keyring)
+  if (!file.exists(file)) {
+    stop("The '", keyring, "' keyring does not exists, create it first!")
+  }
   private$unset_keyring_pass(keyring)
   invisible(self)
 }
 
 b_file_keyring_unlock <- function(self, private, keyring, password) {
-
   file <- private$keyring_file(keyring)
 
   if (!file.exists(file)) {
@@ -291,7 +347,6 @@ b_file_keyring_unlock <- function(self, private, keyring, password) {
 }
 
 b_file_keyring_is_locked <- function(self, private, keyring) {
-
   private$keyring_autocreate(keyring)
 
   keyring <- keyring %||% private$keyring
@@ -304,7 +359,8 @@ b_file_keyring_is_locked <- function(self, private, keyring) {
   if (!file.exists(file_name) || !private$is_set_keyring_pass(keyring)) {
     TRUE
   } else {
-    tryCatch({
+    tryCatch(
+      {
         cached <- private$get_cache(keyring)
         b_file_secret_decrypt(
           cached$check,
@@ -314,10 +370,7 @@ b_file_keyring_is_locked <- function(self, private, keyring) {
         FALSE
       },
       error = function(e) {
-        if(conditionMessage(e) == "Failed to decrypt")
-          TRUE
-        else
-          stop(e)
+        if (conditionMessage(e) == "Failed to decrypt") TRUE else stop(e)
       }
     )
   }
@@ -326,11 +379,17 @@ b_file_keyring_is_locked <- function(self, private, keyring) {
 b_file_keyring_list <- function(self, private) {
   kr_dir <- dirname(private$keyring_file(NULL))
   files <- dir(kr_dir, pattern = "\\.keyring$", full.names = TRUE)
-  names <- sub("\\.keyring", "",  basename(files))
+  names <- sub("\\.keyring", "", basename(files))
   num_secrets <- vapply(
-    files, function(f) length(yaml::yaml.load_file(f)$items), integer(1))
+    files,
+    function(f) length(yaml::yaml.load_file(f)$items),
+    integer(1)
+  )
   locked <- vapply(
-    names, function(k) self$keyring_is_locked(keyring = k), logical(1))
+    names,
+    function(k) self$keyring_is_locked(keyring = k),
+    logical(1)
+  )
   data.frame(
     keyring = unname(names),
     num_secrets = unname(num_secrets),
@@ -351,28 +410,41 @@ b_file_keyring_set_default <- function(self, private, keyring) {
 ## --------------------------------------------------------------------
 ## Private
 
-b__file_keyring_create_direct <- function(self, private, keyring, password, prompt) {
-
+b__file_keyring_create_direct <- function(
+  self,
+  private,
+  keyring,
+  password,
+  prompt
+) {
   keyring <- keyring %||% private$keyring
   prompt <- prompt %||% "Keyring password: "
   file_name <- private$keyring_file(keyring)
 
   if (file.exists(file_name)) {
-    confirmation(paste("are you sure you want to overwrite", file_name,
-                       "(type `yes` if so)"), "yes")
+    confirmation(
+      paste(
+        "are you sure you want to overwrite",
+        file_name,
+        "(type `yes` if so)"
+      ),
+      "yes"
+    )
   }
 
   password <- password %||% get_pass(prompt)
+  if (is.null(password)) stop("Aborted creating keyring")
 
   ## File need to exist for $set_keyring_pass() ...
   dir.create(dirname(file_name), recursive = TRUE, showWarnings = FALSE)
   cat("", file = file_name)
   key <- private$set_keyring_pass(password, keyring)
 
-  with_lock(file_name,
+  with_lock(
+    file_name,
     private$keyring_write_file(
       keyring,
-      nonce = sodium::random(24L),
+      nonce = sodium_random(24L),
       items = list(),
       key = key
     )
@@ -383,13 +455,14 @@ b__file_keyring_create_direct <- function(self, private, keyring, password, prom
 
 b__file_keyring_file <- function(self, private, keyring) {
   keyring <- keyring %||% private$keyring
-  keyring_dir <- getOption("keyring_file_dir",
-                           rappdirs::user_config_dir("r-keyring"))
+  keyring_dir <- getOption(
+    "keyring_file_dir",
+    user_config_dir("r-keyring")
+  )
   file.path(keyring_dir, paste0(keyring, ".keyring"))
 }
 
 b__file_keyring_read_file <- function(self, private, keyring) {
-
   keyring <- keyring %||% private$keyring
   file_name <- private$keyring_file(keyring)
 
@@ -407,16 +480,21 @@ b__file_keyring_read_file <- function(self, private, keyring) {
   )
 
   list(
-    nonce = sodium::hex2bin(yml[["keyring_info"]][["nonce"]]),
+    nonce = sodium_hex2bin(yml[["keyring_info"]][["nonce"]]),
     items = lapply(yml[["items"]], b__file_validate_item),
     check = yml[["keyring_info"]][["integrity_check"]],
     stamp = stamp
   )
 }
 
-b__file_keyring_write_file <- function(self, private, keyring, nonce, items,
-  key) {
-
+b__file_keyring_write_file <- function(
+  self,
+  private,
+  keyring,
+  nonce,
+  items,
+  key
+) {
   keyring <- keyring %||% private$keyring
   file_name <- private$keyring_file(keyring)
   nonce <- nonce %||% private$get_cache(keyring)$nonce
@@ -427,7 +505,7 @@ b__file_keyring_write_file <- function(self, private, keyring, nonce, items,
       list(
         keyring_info = list(
           keyring_version = as.character(getNamespaceVersion(.packageName)),
-          nonce = sodium::bin2hex(nonce),
+          nonce = sodium_bin2hex(nonce),
           integrity_check = b_file_secret_encrypt(
             paste(sample(letters, 22L, replace = TRUE), collapse = ""),
             nonce,
@@ -435,7 +513,7 @@ b__file_keyring_write_file <- function(self, private, keyring, nonce, items,
           )
         ),
         items = items %||% private$get_cache(keyring)$items
-       ),
+      ),
       file_name
     )
   )
@@ -444,7 +522,6 @@ b__file_keyring_write_file <- function(self, private, keyring, nonce, items,
 }
 
 b__file_get_keyring_pass <- function(self, private, keyring) {
-
   kr_env <- b_file_keyring_env(private$keyring_file(keyring))
 
   if (is.null(kr_env$key)) {
@@ -459,7 +536,6 @@ b__file_get_keyring_pass <- function(self, private, keyring) {
 }
 
 b__file_unset_keyring_pass <- function(self, private, keyring) {
-
   kr_env <- b_file_keyring_env(private$keyring_file(keyring))
 
   kr_env$key <- NULL
@@ -473,10 +549,10 @@ b__file_is_set_keyring_pass <- function(self, private, keyring) {
 
 
 b__file_set_keyring_pass <- function(self, private, key, keyring) {
-
   key <- key %||% get_pass("Keyring password: ")
+  if (is.null(key)) stop("Aborted setting keyring password")
   assert_that(is_string(key))
-  key <- sodium::hash(charToRaw(key))
+  key <- sodium_hash(charToRaw(key))
 
   kr_env <- b_file_keyring_env(private$keyring_file(keyring))
 
@@ -484,7 +560,6 @@ b__file_set_keyring_pass <- function(self, private, key, keyring) {
 }
 
 b__file_update_cache <- function(self, private, keyring, nonce, check, items) {
-
   kr_env <- b_file_keyring_env(private$keyring_file(keyring))
 
   kr <- private$keyring_read_file(keyring)
@@ -505,12 +580,15 @@ b__file_update_cache <- function(self, private, keyring, nonce, check, items) {
 }
 
 b__file_get_cache <- function(self, private, keyring) {
-
   keyring_file <- private$keyring_file(keyring)
   kr_env <- b_file_keyring_env(keyring_file)
 
-  if (is.null(kr_env$nonce) || is.null(kr_env$stamp) || is.na(kr_env$stamp) ||
-      file_stamp(keyring_file) != kr_env$stamp) {
+  if (
+    is.null(kr_env$nonce) ||
+      is.null(kr_env$stamp) ||
+      is.na(kr_env$stamp) ||
+      file_stamp(keyring_file) != kr_env$stamp
+  ) {
     kr_env <- private$update_cache(keyring)
   }
 
@@ -520,7 +598,8 @@ b__file_get_cache <- function(self, private, keyring) {
   list(
     nonce = kr_env$nonce,
     items = lapply(kr_env$items, b__file_validate_item),
-    check = kr_env$check)
+    check = kr_env$check
+  )
 }
 
 
@@ -528,21 +607,19 @@ b__file_get_cache <- function(self, private, keyring) {
 ## helper functions
 
 b_file_secret_encrypt <- function(secret, nonce, key) {
-
-  res <- sodium::data_encrypt(
+  res <- sodium_data_encrypt(
     charToRaw(secret),
     key,
     nonce
   )
 
-  b_file_split_string(sodium::bin2hex(res))
+  b_file_split_string(sodium_bin2hex(res))
 }
 
 b_file_secret_decrypt <- function(secret, nonce, key) {
-
   rawToChar(
-    sodium::data_decrypt(
-      sodium::hex2bin(b_file_merge_string(secret)),
+    sodium_data_decrypt(
+      sodium_hex2bin(b_file_merge_string(secret)),
       key,
       nonce
     )
@@ -550,7 +627,6 @@ b_file_secret_decrypt <- function(secret, nonce, key) {
 }
 
 b_file_keyring_env <- function(file_name) {
-
   env_name <- normalizePath(file_name, mustWork = TRUE)
 
   kr_env <- b_file_keyrings[[env_name]]
@@ -563,7 +639,6 @@ b_file_keyring_env <- function(file_name) {
 }
 
 b_file_error <- function(problem, reason = NULL) {
-
   if (is.null(reason)) {
     info <- problem
   } else {
@@ -574,7 +649,6 @@ b_file_error <- function(problem, reason = NULL) {
 }
 
 b__file_validate_item <- function(item) {
-
   assert_that(
     is_list_with_names(item, names = c("service_name", "user_name", "secret")),
     is_string(item[["service_name"]]),
@@ -586,7 +660,6 @@ b__file_validate_item <- function(item) {
 }
 
 b_file_split_string <- function(string, width = 78L) {
-
   assert_that(is_string(string))
 
   paste(
@@ -612,7 +685,8 @@ b__file_keyring_autocreate <- function(self, private, keyring) {
         keyring,
         password = NULL,
         prompt = paste0(
-          "The '", keyring,
+          "The '",
+          keyring,
           "' keyring does not exist, enter a keyring password to create it: "
         )
       )
